////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Lawrence Livermore National Security,
// LLC. Produced at the Lawrence Livermore National Laboratory. 
// Written by Erik Draeger, draeger1@llnl.gov.
//
// LLNL-CODE-432558 All rights reserved.  
//
// This file is part of bigFFT.
//
// bigFFT is distributed under the terms of the GNU General Public License
// as published by the Free Software Foundation, either version 2 of
// the License, or (at your option) any later version.
// See the file COPYING in the root directory of this distribution
// or <http://www.gnu.org/licenses/>.
////////////////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <complex.h>
#include <mpi.h>
#include "bigFFT.h"

int main(int argc, char *argv[]) {

  int npes, mype;
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);
  MPI_Comm_rank(MPI_COMM_WORLD, &mype);

  if (argc != 6)
  {
    printf("usage:  testbigFFT [nx] [ny] [nz] [nprow] [npcol]\n");
    exit;
  }

  // grid dimensions (use smaller if these cause crash)
  const int np0 = atoi(argv[1]);
  const int np1 = atoi(argv[2]);
  const int np2 = atoi(argv[3]);

  // process grid dimensions
  const int nprow = atoi(argv[4]);
  const int npcol = npes/nprow;

  if (mype == 0) 
    printf("testbigFFT started, npes = %i, nprow = %i, npcol = %i, grid = %i x %i x %i\n",npes,nprow,npcol,np0,np1,np2);

#ifdef USE_FFTW
  if (mype == 0) 
    printf("Using FFTW for 1D Fourier transforms\n");
#endif  
#ifdef USE_ESSL
  if (mype == 0) 
    printf("Using ESSL for 1D Fourier transforms\n");
#endif  
  
  BIGFFT_DATA* fftdata;
  fftdata = calloc(1,sizeof(BIGFFT_DATA));
  bigfft_init(fftdata,MPI_COMM_WORLD,np0,np1,np2,nprow,npcol);
  MPI_Barrier(MPI_COMM_WORLD);

  const int np0loc = bigfft_np0loc_bwd(fftdata);
  const int np1locf = bigfft_np1loc_fwd(fftdata);
  const int np1locb = bigfft_np1loc_bwd(fftdata);
  const int np2loc = bigfft_np2loc_fwd(fftdata);

  int size1 = np0*np1locf*np2loc;
  int size2 = 2*np2*np1locb*np0loc;
  const int rhorsize = (size1 > size2 ? size1 : size2);
  double rhor[rhorsize];

  if (mype == 0) 
    printf("testCbigFFT started, npes = %i, nprow = %i, npcol = %i, grid = %i x %i x %i, np0loc=%i, np2loc=%i\n",npes,nprow,npcol,np0,np1,np2,np0loc,np2loc);

  // location of current process on 2D process grid
  int myrow = bigfft_myrow(fftdata);   
  int mycol = bigfft_mycol(fftdata);

  // fill density array with 3D Gaussian
  int ystart[nprow];
  int zstart[npcol];
  ystart[0] = 0;
  for (int i=1; i<nprow; i++) {
    ystart[i] = ystart[i-1] + bigfft_np1loc_row(fftdata,i-1);
  }
  zstart[0] = 0;
  for (int i=1; i<npcol; i++) {
    zstart[i] = zstart[i-1] + bigfft_np2loc_col(fftdata,i-1);
  }

  int i,j,k;
  for (k=0; k<np2loc; k++)
    for (j=0; j<np1locf; j++)
      for (i=0; i<np0; i++) {
        int x = 1 + i;
        int y = 1 + j + ystart[myrow];
        int z = 1 + k + zstart[mycol];
        double half = 0.5*np0+4.0;
        double rsq = ((double)x - half)*((double)x - half)/3200. + ((double)y - half)*((double)y - half)/4400. + ((double)z - half)*((double)z - half)/6500.;
        rhor[i + j*np0 + k*np0*np1locf] = exp( -1.*rsq );
      }  

  //for (int i=0; i<np0*np1locf*np2loc; i++)
  //  rhor[i] = 4148194.3*random();


  //save initial values for comparison 
  double rhor2[rhorsize];
  for (i=0; i<rhorsize; i++) {
    const double v = rhor[i];
    rhor2[i] = v;
  }

  // forward fft
  const double scale = 1./(np0*np1*np2);
  MPI_Barrier(MPI_COMM_WORLD);
  double tstart = MPI_Wtime();
  int fwd_status = bigfft_forward(fftdata,rhor,scale);
  MPI_Barrier(MPI_COMM_WORLD);
  double tend = MPI_Wtime();
  if (mype == 0) 
    printf("bigFFT.forward time:  mype = %i, fwd time = %0.7f sec\n",mype,tend-tstart);

  // backward fft
  MPI_Barrier(MPI_COMM_WORLD);
  double tstart2 = MPI_Wtime();
  int bwd_status = bigfft_backward(fftdata,rhor,1.0);
  MPI_Barrier(MPI_COMM_WORLD);
  double tend2 = MPI_Wtime();
  if (mype == 0) 
    printf("bigFFT.backward time:  mype = %i, bwd time = %0.7f sec\n",mype,tend2-tstart2);


  // debug:  calculate difference between result of forward+backward transform and original
  double maxdiffsq = 0.0;
  for (k=0; k<np2loc; k++)
    for (j=0; j<np1locf; j++)
      for (i=0; i<np0; i++) {
        double v = (rhor2[i + j*np0 + k*np0*np1locf] - rhor[i + j*np0 + k*np0*np1locf]);
        double diffsq = v*v;
        if (diffsq > maxdiffsq) maxdiffsq = diffsq;
      }
  if (mype == 0) 
    printf("bigFFT.diffsq, mype = %i, max diff = %0.10f\n",mype,sqrt(maxdiffsq));

  if (maxdiffsq > 1.E-4) {
    printf("WARNING:  bigFFT.diffsq, mype = %i, max diff = %0.10f\n",mype,sqrt(maxdiffsq));
  }
  
  bigfft_cleanup(fftdata);

  MPI_Finalize();

}
